#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#if defined(_MSC_VER)
typedef unsigned char uint8_t;
typedef unsigned short uint16_t;
typedef signed short int16_t;
typedef unsigned int uint32_t;
typedef signed int int32_t;
#else
#include <inttypes.h>
#endif

uint16_t csum;
uint8_t b;
uint8_t* bmppal;
uint8_t* bmp;
uint8_t* gifimg, *gifend;
uint8_t* out;
uint8_t* gifheader;
uint16_t w;
uint16_t ww;
uint16_t h;
uint16_t numcol;
uint16_t currentcode;
int gifbpp, x, y, i, lzstart, lzend;
FILE* of;
int32_t* lengths;
int32_t l0[] = {0};
int32_t l1[] = {18,8760,4679,10029,11068,10742,10655,11549,0};
int32_t l2[] = {70,175,21,10,126894,2,112828,0};
int32_t l3[] = {4,19438,23230,31160,21736,39362,46,24,0};
int32_t l4[] = {10,9557,10258,14612,10530,9851,3829,6889,0};
int32_t l5[] = {1800255,2394049,0};
int32_t l6[] = {295,65212,29,0};
int32_t l7[] = {1778,638222,0};
int32_t l8[] = {73,76,108,113,37,74,0};
uint8_t byte;
uint8_t bits;
uint8_t* pos, *startpos;
int32_t dist;
uint8_t dw[4096][8192];
uint16_t dl[4096];

void writebits(int code, int length) {
    while(length-->0){
        byte |= (code&1)<<bits;
        code >>= 1;
        bits++;

        if (bits == 8) {
            *pos = byte;
            pos++;
            byte=0;
            bits=0;
        }
    }
}

void lzw(uint32_t length) {
    uint8_t w[8192];
    uint16_t l;
    uint16_t i, codesize, limit;
    int code;

    for (i = 0; i < 1<<gifbpp; i++) {
        dw[i][0] = i;
        dl[i] = 1;
    }
    currentcode = (1<<gifbpp) + 1;
    codesize = gifbpp + 1;

    dl[currentcode-1] = 0;
    dl[currentcode] = 0;
    dl[currentcode+1] = 0;

    l = 0;

    while (length-->0) {
        uint8_t c = *gifimg;
        gifimg++;
        w[l] = c;
        l++;

        code = find(w, l);
        if (code<0) {
            writebits(find(w,l-1), codesize);
            if(currentcode<4095){
                currentcode++;
                memcpy(dw[currentcode], w, l);
                dl[currentcode] = l;
                if (currentcode != 4096 && currentcode == (1 << codesize)) {
                    codesize++;
                }
            }

            w[0] = c;
            l = 1;
        }
    }

    writebits(find(w,l), codesize);

    if(currentcode<4095){
        currentcode++;
        if (currentcode != 4096 && currentcode == (1 << codesize)) {
            codesize++;
        }
    }
    writebits((1<<gifbpp) +(gifimg==gifend?1:0), codesize);
}

int find(uint8_t w[8192], uint16_t l) {
    int i;
    for (i = 0; i <= currentcode; i++) {
        if(l == dl[i] && !memcmp(w, dw[i], l))
            return i;
    }

    return -1;
}

int main(int argc, char** argv)
{
    FILE* fp = fopen(argv[1], "rb");
    uint16_t* header = malloc(0x36);

    fread(header,6,9,fp);

    w = *(header+9);
    ww = (w+3)&~3;
    h = *(header+11);
    numcol = *(header+23);
    if(!numcol)numcol=256;

    gifbpp = 2;
    while (1<<gifbpp < numcol)
        gifbpp++;

    bmppal = malloc(numcol*4);
    fread(bmppal, numcol*4,1,fp);

    bmp = malloc(ww*h);
    fread(bmp,ww*h,1,fp);

    csum = 0;
    gifimg = malloc(w*h);
    gifend = gifimg + w*h;
    for(y = 0; y < h; y++) {
        for(x = 0; x < w; x++) {
            b = *(bmp+ww*(h-1-y)+x);
            csum = csum << 1 ^ b;
            *(gifimg+y*w+x) = b;
        }
    }

    switch(csum){
        case 227:
            lengths = l1;
            break;
        case 5267:
            lengths = l2;
            break;
        case 6144:
            lengths = l3;
            break;
        case 53282:
            lengths = l4;
            break;
        case 7:
            lengths = l5;
            break;
        case 10212:
            lengths = l6;
            break;
        case 65535:
            lengths = l7;
            break;
        case 16576:
            lengths = l8;
            break;
        default:
            lengths = l0;
    }

    of = fopen(argv[2], "wb");
    gifheader = "GIF89aXXXX\xf0\0";
    *((uint16_t*) (gifheader+6)) = w;
    *((uint16_t*) (gifheader+8)) = h;
    gifheader[10] |= gifbpp-1;
    fwrite(gifheader,13,1,of);
    for (i = 0; i < 1<<gifbpp; i++) {
        fputc(i<numcol?bmppal[i*4+2]:0,of);
        fputc(i<numcol?bmppal[i*4+1]:0,of);
        fputc(i<numcol?bmppal[i*4+0]:0,of);
    }

    *( (int32_t*) (gifheader+2)) = 0;
    gifheader[1] = 0x2c;
    gifheader[10] = 0;
    gifheader[11] = gifbpp;
    fwrite(gifheader+1,11,1,of);


    pos = malloc(2048*2048*2);
    startpos = pos;

    writebits(1 << gifbpp, gifbpp + 1); // clr code

    if(*lengths){
        while (*lengths) {
            lzw(*lengths);
            lengths++;
        }
    } else {
        lzw(w*h);
    }

    *pos = byte;
    pos++;

    while(pos-startpos) {
        dist = pos-startpos;
        if (dist > 0xff) dist = 0xff;
        fputc((uint8_t) dist, of);
        fwrite(startpos,dist,1,of);
        startpos += dist;
    }
    fputc(0,of);
    fputc(59,of);
    fclose(of);
    return 0;
}
